/*
 * title:       Stock Utility Routines
 *
 * author:      Craig E. Ward
 *
 * envionment:  GCC 3.1
 */
#include <stdlib.h>
#include <string.h>

#include "mytypes.h"
#include "util.h"
#include "stocks.h"

/* static functions */

/*
 * Allocate and initalize to all zeros memory for a portfolio entry
 */
static struct portfolio_entry *new_entry(void)
{
  struct portfolio_entry *new_stock = 
    (struct portfolio_entry *)malloc(sizeof(struct portfolio_entry));

  memset(new_stock, '\0', sizeof(struct portfolio_entry));
  new_stock->astock = (struct stock *)malloc(sizeof(struct stock));
  memset(new_stock->astock, '\0', sizeof(struct stock));

  return new_stock;
}

/*
 * Free the allocations for a portfolio_entry structure.
 * NOTE: It should already have been removed from the linked-list!
 */
static void free_entry(struct portfolio_entry *entry)
{
  free(entry->astock);          /* release the stock structure */
  free(entry);                  /* release the portfolio structure */
  return;
}

/* public functions */

/*
 * Given a symbol lookup the index for that name
 */
int find_stock_by_name(const char *symbol)
{
  log(__FILE__ ".find_stock_by_name: %s", symbol);
  int i = 0;

  for (i = 0; i < NUMBER_SYMBOLS; i++) {
    if ((strcmp(symbol, symbols[i])) == 0) {
      break;
    }
  }
  log(__FILE__ ".find_stock_by_name: returning %d for %s", i, symbol);
  return i;
}

/*
 * Add a stock to linked-list of this portofolio. Linked-list is NOT sorted.
 */
struct portfolio_entry *add_stock(struct portfolio_entry *portfolio,
               int symbol_code,
               int shares,
               double price)
{
  log(__FILE__ ".add_stock: code=%d, shares=%d, unit price=%.0f",
      symbol_code,
      shares,
      price);
  struct portfolio_entry *top = portfolio;

  if (top == NULL) {
    log(__FILE__ ".add_stock: first stock!");
    top = new_entry();
    top->shares = shares;
    strcpy(top->astock->symbol, symbols[symbol_code]);
    top->astock->code = symbol_code;
    top->astock->unitvalue = price;
  } else {
    log(__FILE__ ".add_stock: adding new stock to portfolio");
    struct portfolio_entry *current = portfolio;
    while (current->next != NULL) {
      current = current->next;
    }
    current->next = new_entry();
    current->next->shares = shares;
    strcpy(current->next->astock->symbol, symbols[symbol_code]);
    current->next->astock->code = symbol_code;
    current->next->astock->unitvalue = price;
  }
  log(__FILE__ ".add_stock: return");
  return top;
}
/*
 * Step through the linked-list looking for the delisted "victim" stock issue.
 */
struct portfolio_entry *remove_stock(struct portfolio_entry * portfolio,
                  int symbol_code)
{
  log(__FILE__ ".remove_stock: code=%d", symbol_code);

  struct portfolio_entry *top = portfolio; /* remember the top of list */
  struct portfolio_entry *current = portfolio; /* for moving through list */
  struct portfolio_entry *previous = current; /* before the victim */
  struct portfolio_entry *victim = NULL; /* the victim */
  
  /* Step through list looking for the stock to remove
     ------------------------------------------------- */
  while (current != NULL) {
    log(__FILE__ ".remove_stock: compare %d to %d",
        current->astock->code, symbol_code);
    if (current->astock->code == symbol_code) {
      log(__FILE__ ".remove_stock: found match");
      victim = current;
      break;
    }
    previous = current;
    current = current->next;
  }

  /* If we found a victim, see if it's the head otherwise, assign its next
     pointer to the previous next pointer and release the memory
     --------------------------------------------------------------------- */
  if (victim != NULL) {
    if (victim->astock->code == top->astock->code) {
      log(__FILE__ ".remove_stock: removing the head!");
      top = top->next;
    } else {
      log(__FILE__ ".remove_stock: relinking after head!");
      previous->next = victim->next;
    }
    free_entry(victim);         /* return the memor */
  } else {
    log(__FILE__ ".remove_stock: code=%d NOT FOUND", symbol_code);
  }

  log(__FILE__ ".remove_stock: return");
  return top;
}
/*
 * Update the share count of a stock. To remove shares, send a negative
 * number. Return the new value.
 */
int update_stock_count(struct portfolio_entry *portfolio,
                        int symbol_code,
                        int share_change)
{
  log(__FILE__ ".update_stock_count: code=%d, change=%d",
      symbol_code, share_change);
  struct portfolio_entry *current = portfolio;
  int count = 0;
  while (current != NULL) {
    if (current->astock->code == symbol_code) {
      break;
    }
    current = current->next;
  }

  if (current != NULL) {
    current->shares = current->shares + share_change;
    if (current->shares < 0) current->shares = 0;
    count = current->shares;
  } else {
    log(__FILE__ ".update_stock_count: code=%d NOT FOUND", symbol_code);
  }

  log(__FILE__ ".update_stock_count: return");
  return count;
}

/*
 * Update the unitvalue of a stock with a new price
 */
void update_stock_price(struct portfolio_entry *portfolio,
                        int symbol_code,
                        double new_price)
{
  log(__FILE__ ".update_stock_price: code=%d, change=%.0f",
      symbol_code, new_price);
  struct portfolio_entry *current = portfolio;

  while (current != NULL) {
    if (current->astock->code == symbol_code) {
      break;
    }
    current = current->next;
  }

  if (current != NULL) {
    current->astock->unitvalue = new_price;
  } else {
    log(__FILE__ ".update_stock_price: code=%d NOT FOUND", symbol_code);
  }

  log(__FILE__ ".update_stock_price: return");
  return;
}

/*
 * Scan the portfolio to see if we own any of a particular stock
 */
bool_t is_owned(struct portfolio_entry *portfolio,
                int symbol_code)
{
  log(__FILE__ ".is_owned: code=%d", symbol_code);
  bool_t found = NO;
  struct portfolio_entry *current = portfolio;

  while (current != NULL) {
    if (current->astock->code == symbol_code) {
      found = YES;
      break;
    }
    current = current->next;
  }

  log(__FILE__ ".is_owned: returning: %s",
      (found == YES ? "YES" : "NO"));
  return found;
}

/*
 * Market support functions
 */

/*
 * Search the listings for the indicated stock and return a pointer to it.
 */
struct listing *get_stock_listing(struct listing *listings,
                                  int symbol_code)
{
  log(__FILE__ ".get_stock_listing: code=%d", symbol_code);
  struct listing *current = listings;

  while (current->next != NULL) {
    if (current->astock->code == symbol_code)
      break;
    current = current->next;
  }
  log(__FILE__ ".get_stock_current: found \"%s\"", current->astock->symbol);
  return current;
}
